/*
 * Copyright 2008 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _CONDOR_ALLOW_OPEN
#include "condor_common.h"

#include "JobObject.h"

#include "ArgsJobGetAd.h"
#include "ArgsJobSetAttribute.h"
#include "ArgsJobHold.h"
#include "ArgsJobRelease.h"
#include "ArgsJobRemove.h"
#include "ArgsJobFetch.h"

#include "MgmtScheddPlugin.h"

#include "set_user_priv_from_ad.h"
#include "../condor_schedd.V6/scheduler.h"

#include "stdint.h"

#include "JobUtils.h"

#include "MgmtConversionMacros.h"


// Global from the condor_schedd, it's name
extern char * Name;

extern MgmtScheddPlugin *scheddPluginInstance;


using namespace mrg::grid;

using namespace qpid::management;
using namespace qmf::mrg::grid;

JobObject::JobObject(ManagementAgent *agent,
					 SchedulerObject *_scheduler,
					 uint32_t _clusterId,
					 uint32_t _procId,
					 const SubmitterObject *submitter):
	clusterId(_clusterId),
	procId(_procId)
{
	uint64_t id = 0;
	MyString name;

	mgmtObject = new Job(agent, this, _scheduler);

	mgmtObject->set_submitterRef(submitter->GetManagementObject()->getObjectId());

		// The id space starting with 0 as the upper 32 bits is
		// reserved, the rest of the id space is for jobs, no job can
		// have a clusterId of 0
	id = (((uint64_t) clusterId) << 32) + procId;
	agent->addObject(mgmtObject, id);
}


JobObject::~JobObject()
{
	if (mgmtObject) {
		mgmtObject->resourceDestroy();
	}
}


void
JobObject::update(const ClassAd &ad)
{
	MGMT_DECLARATIONS;

	OPT_STRING(AccountingGroup);
	OPT_STRING(Args);
	INTEGER(ClusterId);
	STRING(Cmd);
	OPT_STRING(ConcurrencyLimits);
	OPT_STRING(CustomGroup);
	OPT_STRING(CustomId);
	OPT_INTEGER(CustomPriority);
	STRING(GlobalJobId);
	STRING(In);
	STRING(Iwd);
	INTEGER(JobStatus);
	INTEGER(JobUniverse);
	OPT_STRING(Note);
	STRING(Out);
	STRING(Owner);
	STRING(User);
	INTEGER(ProcId);
	TIME_INTEGER(QDate);
//	EXPR(Requirements);
	OPT_STRING(Title);
	OPT_STRING(UserLog);

	OPT_STRING(HoldReason);

	OPT_STRING(DAGNodeName);
	OPT_STRING(DAGParentNodeNames);
	OPT_INTEGER(DAGManJobId);

		// Provide a default for the CustomGroup
	if (!mgmtObject->isSet_CustomGroup()) {
		std::ostringstream id_stream;

			// If we are a DAG node, we default to our DAG group
		if (mgmtObject->isSet_DAGManJobId()) {
			PROC_ID id;
			JobObject *job;

				// The id of the DAGMan job that spawned us
			id.cluster = mgmtObject->get_DAGManJobId();
			id.proc = 0;

				// We *MUST* ALWAYS be created AFTER the DAGMan
				// job that spawned us. This means processing of
				// the dirty jobs should always be in increasing
				// sequential order.
			if (-1 == scheddPluginInstance->getJobTable()->lookup(id, job)) {
					// This can also happen if the DAGMan job was
					// removed, and we remained, which should not
					// happen, but could. In such a case we are
					// orphaned, and we'll make a guess at the proper
					// CustomGroup. We'll be wrong if the DAGMan job
					// didn't use the default, Here, it is better to
					// be wrong than to fail entirely, which is the
					// alternative.
				id_stream << Name << "#" << mgmtObject->get_DAGManJobId();
			} else {
				id_stream <<
					((Job *) job->GetManagementObject())->get_CustomGroup();
			}
		} else {
			id_stream << Name << "#" << mgmtObject->get_ClusterId();
		}

		mgmtObject->set_CustomGroup(id_stream.str());
	}

		// Provide a default for the CustomId
	if (!mgmtObject->isSet_CustomId()) {
		std::ostringstream id;

		id << mgmtObject->get_ClusterId()
		   << "."
		   << mgmtObject->get_ProcId();

		mgmtObject->set_CustomId(id.str());
	}
}


ManagementObject *
JobObject::GetManagementObject(void) const
{
	return mgmtObject;
}


Manageable::status_t
JobObject::GetAd(qpid::framing::FieldTable &fieldTable, std::string &/*text*/)
{
	if (!PopulateFieldTableFromJobAd((int) clusterId,
									 (int) procId,
									 fieldTable)) {
		return STATUS_UNKNOWN_OBJECT;
	}

	return STATUS_OK;
}


Manageable::status_t
JobObject::SetAttribute(std::string name, std::string value, std::string &text)
{
		// All values are strings in the eyes of
		// ::SetAttribute. Their type is inferred when read from
		// the ClassAd log. It is important that the incoming
		// value is properly quoted to differentiate between EXPR
		// and STRING.
	if (::SetAttribute((int) clusterId,
					   (int) procId,
					   name.c_str(),
					   value.c_str())) {
		text = "Failed to set attribute " + name + " to " + value;
		return STATUS_USER + 2;
	}

	return STATUS_OK;
}


Manageable::status_t
JobObject::Hold(std::string &reason, std::string &text)
{
	if (!holdJob(clusterId,
				 procId,
				 reason.c_str(),
				 true, // Always perform this action within a transaction
				 true, // Always notify the shadow of the hold
				 false, // Do not email the user about this action
				 false, // Do not email admin about this action
				 false // This is not a system related (internal) hold
				 )) {
		text = "Failed to hold job";
		return STATUS_USER + 1;
	}

	return STATUS_OK;
}


Manageable::status_t
JobObject::Release(std::string &reason, std::string &text)
{
	if (!releaseJob(clusterId,
					procId,
					reason.c_str(),
					true, // Always perform this action within a transaction
					false, // Do not email the user about this action
					false // Do not email admin about this action
					)) {
		text = "Failed to release job";
		return STATUS_USER + 1;
	}

	return STATUS_OK;
}


Manageable::status_t
JobObject::Remove(std::string &reason, std::string &text)
{
	if (!abortJob(clusterId,
				  procId,
				  reason.c_str(),
				  true // Always perform within a transaction
				  )) {
		text = "Failed to remove job";
		return STATUS_USER + 1;
	}

	return STATUS_OK;
}


Manageable::status_t
JobObject::Fetch(std::string &file,
				 int32_t start,
				 int32_t end,
				 std::string &data,
				 std::string &text)
{
	ClassAd *ad;
	priv_state prev_priv_state;
	int fd = -1;
	int count;
	int length;
	int whence;
	char *buffer;
	Manageable::status_t status;

		// start >= 0, end >= 0 :: lseek(start, SEEK_SET), read(end - start)
		//  end < start :: error, attempt to read backwards
		// start >= 0, end < 0 :: error, don't know length
		// start < 0, end > 0 :: attempt to read off end of file, end = 0
		// start < 0, end <= 0 :: lseek(start, SEEK_END), read(abs(start - end))
		//  end < start :: error, attempt to read backwards

	if ((start >= 0 && end >= 0 && end < start) ||
		(start >= 0 && end < 0) ||
		(start < 0 && end <= 0 && end < start)) {
		text = "Invalid start and end values";
		return STATUS_USER + 10;
	}

		// Instead of reading off the end of the file, read to the
		// end of it
	if (start < 0 && end > 0) {
		end = 0;
	}

	if (start >= 0) {
		whence = SEEK_SET;
		length = end - start;
	} else {
		whence = SEEK_END;
		length = abs(start - end);
	}

		// XXX: Sanity check that length isn't too big?

	buffer = new char[length + 1];

	if (NULL == (ad = GetJobAd((int) clusterId, (int) procId, false))) {
		dprintf(D_ALWAYS,
				"Fetch method called on %d.%d, which does not exist\n",
				clusterId, procId);
		return STATUS_UNKNOWN_OBJECT;
	}

	prev_priv_state = set_user_priv_from_ad(*ad);

	if (-1 != (fd = safe_open_wrapper(file.c_str(),
									  O_RDONLY | _O_BINARY,
									  0))) {
			// If we are seeking from the end of the file, it is
			// possible that we will try to seek off the front of the
			// file. To avoid this, we check the file's size and set
			// the max length that we are able to seek
		if (SEEK_END == whence) {
			struct stat buf;
			if (-1 != fstat(fd, &buf)) {
				if (buf.st_size < abs(start)) {
					start = -buf.st_size;
				}
			} // if fstat fails, we just continue having made an effort
		}

		if (-1 != lseek(fd, start, whence)) {				
			if (-1 == (count = full_read(fd, buffer, length))) {
				text = "Failed to read from " + file;
				status = STATUS_USER + 3;
			} else {
					// Terminate the string.
				buffer[count] = '\0';

				data = buffer;
				status = STATUS_OK;
			}

			close(fd); // assume closed on failure?
		} else {
			text = "Failed to seek in " + file;
			status = STATUS_USER + 2;
		}
	} else {
		text = "Failed to open " + file;
		status = STATUS_USER + 1;
	}

	set_priv(prev_priv_state);

	delete []buffer;

	return status;
}


Manageable::status_t
JobObject::ManagementMethod(uint32_t methodId, Args &args, std::string &text)
{
	switch (methodId) {
	case Job::METHOD_GETAD:
		return GetAd(((ArgsJobGetAd &) args).o_JobAd, text);
	case Job::METHOD_SETATTRIBUTE:
		return SetAttribute(((ArgsJobSetAttribute &) args).i_Name,
							((ArgsJobSetAttribute &) args).i_Value,
							text);
	case Job::METHOD_HOLD:
		return Hold(((ArgsJobHold &) args).i_Reason, text);
	case Job::METHOD_RELEASE:
		return Release(((ArgsJobRelease &) args).i_Reason, text);
	case Job::METHOD_REMOVE:
		return Remove(((ArgsJobRemove &) args).i_Reason, text);
	case Job::METHOD_FETCH:
		return Fetch(((ArgsJobFetch &) args).i_File,
					 ((ArgsJobFetch &) args).i_Start,
					 ((ArgsJobFetch &) args).i_End,
					 ((ArgsJobFetch &) args).o_Data,
					 text);
	}

	return STATUS_NOT_IMPLEMENTED;
}

